with Ada.Characters.Handling,
     Ada.Calendar,
     Ada.Strings.Fixed,
     Claw.Time,
     Sorts,
     WORM_Str;
package body Cookie is

  Mon : constant array (Ada.Calendar.Month_Number) of String(1 .. 3)
    := ("Jan", "Feb", "Mar", "Apr", "May", "Jun",
        "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");

  function Full_Date(Source : String) return Ada.Calendar.Time is
  -- Cookie dates are of the form:  Sun, 06-Nov-1994 08:49:37 GMT
  -- Like RFC1123 except for the '-' separators, and the time
  -- zone Must be "GMT".
  -- We'll be somewhat tolerant, since practice doesn't follow the spec.
  -- Raises Constraint_Error if problems.
    use Ada.Calendar;
    S  : String(1 .. Source'Length) := Source;
    D  : Day_Number := Day_Number'First;
    M  : Month_Number := Month_Number'First;
    Y4 : Year_Number := Year_Number'First;
    Hh : Integer range 0 .. 23 := 0;
    Mm : Integer range 0 .. 59 := 0;
    Ss : Integer range 0 .. 59 := 0;
    si : Natural;
  begin
    if S(4) /= ',' then  -- Full day-of-week word, not 3 letter abbrev?
      si := Ada.Strings.Fixed.Index(S,",");
      if si not in 7 .. 10 then raise Constraint_Error;end if;
      S := S(si-3 .. S'last) & (1 .. si-4 => ' ');
    end if;
    if S(8) /= '-' and S(8) /= ' ' then  -- allow blank instead of -
      raise Constraint_Error;
    end if;
    if S(12) /= '-' and S(12) /= ' ' then
      raise Constraint_Error;
    end if;
    D := Ada.Calendar.Day_Number'Value(S(6 .. 7));
    for I in Month_Number loop
      if Mon(I) = S(9 .. 11) then
        M := I;
        exit;
      end if;
      if I = Month_Number'Last then
        raise Constraint_Error;
      end if;
    end loop;
    Y4 := Ada.Calendar.Year_Number'Value(S(13 .. 16));
    Hh := Integer'Value(S(18 .. 19));
    Mm := Integer'Value(S(21 .. 22));
    Ss := Integer'Value(S(24 .. 25));
    return Time_Of(Y4, M, D, Day_Duration(Hh) * 3600
                             + Day_Duration(Mm) * 60
                             + Day_Duration(Ss));
  end Full_Date;


  procedure Put_String(Host : in String;
                       Path : in String;
                       Text : in String) is
  -- from "Set-Cookie: Text"
    use Ada.Strings.Fixed;
    use Ada.Characters.Handling;
    Secure : Boolean := False;
    Data_A : Natural := Text'first;
    Data_Z, Path_A, Path_Z, Domain_A, Domain_Z : Natural:=0;
    Next_A, Next_Z : Natural;
    Expires : Ada.Calendar.Time := Unknown_Time;
    procedure Find_Next(S : in String) is
    begin
      Next_A := Index_Non_Blank(S);
      if Next_A = 0 then return;end if;
      Next_Z := Index(S(Next_A .. Text'last), ";");
      if Next_Z = 0 then return;end if;
    end Find_Next;
    function Domain_To_Use return String is
    begin
      if Domain_A /= 0 then return Text(Domain_A .. Domain_Z-1);end if;
      return Host(Host'first+2 .. Host'last);  -- w/o //
    end Domain_To_Use;
    function Path_To_Use return String is
    begin
      if Path_A /= 0 then return Text(Path_A .. Path_Z-1);end if;
      return Path;
    end Path_To_Use;
  begin
    if Path'length = 0 or else Path(Path'first) /= '/' then
      Put_String(Host, '/' & Path, Text);
      return;
    end if;
    if Host'length < 2 or else Host(Host'first .. Host'first+1) /= "//" then
      return;  -- illegal host name
    end if;
    -- *_Z -> ';'
    Data_Z := Index(Text, ";");
    if Data_Z = 0 then Data_Z := Text'last+1;end if;
    if Index(Text(Data_A .. Data_Z-1), "=") = 0 then  -- no '=' in name=value
      return;
    end if;
    Next_Z := Data_Z;
    while Next_Z < Text'last loop
      Find_Next(Text(Next_Z+1 .. Text'last));
      exit when Next_A = 0;
      if Next_Z = 0 then
        Next_Z := Text'last+1;
      end if;
      if Next_A+4 <= Next_Z
      and then To_Lower(Text(Next_A .. Next_A+4)) = "path=" then
        Path_A := Next_A+5;
        Path_Z := Next_Z;
      elsif Next_A+6 <= Next_Z
      and then To_Lower(Text(Next_A .. Next_A+6)) = "domain=" then
        Domain_A := Next_A+7;
        Domain_Z := Next_Z;
      elsif Next_A+7 <= Next_Z
      and then To_Lower(Text(Next_A .. Next_A+7)) = "expires=" then
        begin
          Expires := Full_Date(Text(Next_A+8 .. Next_A+36));
        exception
          when Constraint_Error => null;
        end;
      elsif Next_A+5 <= Next_Z
      and then To_Lower(Text(Next_A .. Next_A+5)) = "secure" then
        Secure := True;
      -- else ignore others
      end if;
    end loop;
    -- only hosts that fall within a domain may specify that domain
    if Domain_A /= 0 then
      if Host'length < Domain_Z - Domain_A     -- Domain_Z -> ';'
      or else To_Lower(Host(Host'last-(Domain_Z-1-Domain_A) .. Host'last))
          /= To_Lower(Text(Domain_A .. Domain_Z-1)) then
        return;
      end if;
    end if;
    -- all is well
    Put(Data=>Text(Data_A .. Data_Z-1),
        Path=>Path_To_Use, Domain=>Domain_To_Use,
        Expiration=>Expires,
        Secure=>Secure);
  end Put_String;

  protected Cookie_Jar is
    procedure Put(Data, Path, Domain : in String;
                  Expiration : in Ada.Calendar.Time;
                  Secure : in Boolean);
    function Get(Host : String;
                 Path : String;
                 Secure : Boolean := False) return String;
    function Take_Snapshot return Snapshot_Type;
  private
    Jar : Jar_Contents_Type;
    Count : Natural := 0;
    String_Storage : String_Storage_Type;
  end Cookie_Jar;

  protected body Cookie_Jar is

    procedure Toss_Garbage is
      use WORM_Str;
      use type Ada.Calendar.Time;
      Temp : String_Storage_Type;  -- fill with only strings in actual use
    begin
      for i in 1 .. Count loop
        if Jar(i).Expiration = Unknown_Time
        or else Jar(i).Expiration >= Ada.Calendar.Clock then
          Add(Temp, Get(String_Storage, Jar(i).Data),  Jar(i).Data);
          Add(Temp, Get(String_Storage, Jar(i).Path),  Jar(i).Path);
          Add(Temp, Get(String_Storage, Jar(i).Domain),Jar(i).Domain);
        end if;
      end loop;
      String_Storage := Temp;
    end Toss_Garbage;

    procedure Toss_Oldest is
    begin
      if Count > 0 then
        Jar(Jar'first .. Jar'last-1) := Jar(Jar'first+1 .. Jar'last);
        Count := Count-1;
        Toss_Garbage;
      end if;
    end Toss_Oldest;

    procedure Add(Data, Path, Domain : in String;
                  Expiration : in Ada.Calendar.Time;
                  Secure : in Boolean;
                  Try : in Positive) is
      -- Note: Normally, but not necessarily, called with OK Expiration.
      -- It merely wastes space to keep track of expired cookies since
      -- they'll not be returned by Get anyway, and Toss_Garbage will
      -- get rid of them if space is a problem.
      use Ada.Characters.Handling;
      Temp : Infos;
      function Match(Data1, Data2, Path1, Path2 : String) return Boolean is
        use Ada.Strings.Fixed;
        End_Name1 : Positive := Ada.Strings.Fixed.Index(Data1, "=");
        End_Name2 : Positive := Ada.Strings.Fixed.Index(Data2, "=");
      begin
        if Data1(Data1'first .. End_Name1) /= Data2(Data2'first .. End_Name2) then
          return False;
        end if;
        -- Name parts are the same, check paths
        return Path1 = Path2;
      end Match;
    begin
      -- check if we already have this name & path
      for i in 1 .. Count loop
        if Match(WORM_Str.Get(String_Storage,Jar(i).Data), Data,
                 WORM_Str.Get(String_Storage,Jar(i).Path), Path) then
          -- Name part of Data, and Path, match.  Check if whole thing matches
          if Data /= WORM_Str.Get(String_Storage,Jar(i).Data)
          or else To_Lower(Domain) /= WORM_Str.Get(String_Storage,Jar(i).Domain) then
            -- Data or Domain is different, replace old cookie
            Temp := Jar(i);  -- don't allow Jar(i) to be "in process" modification
            WORM_Str.Add(String_Storage,Data, Temp.Data);
            WORM_Str.Add(String_Storage,To_Lower(Domain), Temp.Domain);
            Temp.Expiration := Expiration;
            Temp.Secure := Secure;
            Jar(i) := Temp;
          end if;
          return;
        end if;
      end loop;
      -- New, add it
      if Count = Jar'last then -- toss oldest to make room
        Toss_Oldest;
      end if;
      WORM_Str.Add(String_Storage,Data,Temp.Data);
      WORM_Str.Add(String_Storage,Path,Temp.Path);
      WORM_Str.Add(String_Storage,To_Lower(Domain),Temp.Domain);
      Temp.Expiration := Expiration;
      Temp.Secure := Secure;
      Count := Count+1;
      Jar(Count) := Temp;
    exception
      when WORM_Str.Overflow =>
        if Try = 1 then
          Toss_Garbage;
          Add(Data, Path, Domain, Expiration, Secure, Try => Try+1); -- try again
        elsif Count > Max_Simultaneous_Cookies / 2 then
          Toss_Oldest;
          Add(Data, Path, Domain, Expiration, Secure, Try => Try+1); -- try again
        -- else this huge cookie won't fit even in 1/2 full jar, ignore it
        end if;
    end Add;

    procedure Put(Data, Path, Domain : in String;
                  Expiration : in Ada.Calendar.Time;
                  Secure : in Boolean) is
      use type Ada.Calendar.Time;
    begin
      if Expiration = Unknown_Time
      or else Expiration >= Ada.Calendar.Clock then
        Add(Data, Path, Domain, Expiration, Secure, Try=>1);
      end if;
    end Put;

    function Get(Host : String;
                 Path : String;
                 Secure : Boolean := False) return String is
      use Ada.Characters.Handling;
      use type Ada.Calendar.Time;
      LC_Host : constant String := To_Lower(Host);
      The_Path : constant String := Path;
      Result : String(1 .. Max_Cookie_String);
      Last : Natural := 0;
      type Pieces is record     -- for sorting Result chunks by Path length
        First, Last : Natural;
        Path_Length : Natural;
      end record;
      type Piece_Lists is array(integer range <>) of Pieces;
      function "<"(Left, Right : Pieces) return Boolean is
      begin -- we want longer, ie more specific, paths first
        return Left.Path_Length > Right.Path_Length;
      end "<";
      procedure Sort_Pieces is new Sorts.List_Merge(Item=>Pieces,
                                     List_Type=>Piece_Lists);
      Piece_List : Piece_Lists(1 .. Count);
      Piece_Count : Natural := 0;
      function Sorted_Result return String is
          Temp : String(1 .. Last+2*Piece_Count-1);
          di : Natural := 1;
          Length : Natural;
      begin
        if Piece_Count > 1 then
          Sort_Pieces(Piece_List(1 .. Piece_Count));
        end if;
        for i in 1 .. Piece_Count loop
          if i > Piece_List'first then
            Temp(di) := ';';
            Temp(di+1) := ' ';
            di := di+2;
          end if;
          Length := Piece_List(i).Last - Piece_List(i).First+1;
          Temp(di .. di+Length-1)
            := Result(Piece_List(i).First .. Piece_List(i).Last);
          di := di+Length;
        end loop;
        return Temp(1 .. di-1);
      end Sorted_Result;

      procedure Check_and_Build_Result(X : in Infos) is
        Data : constant String := WORM_Str.Get(String_Storage,X.Data);
        Path : constant String := WORM_Str.Get(String_Storage,X.Path);
        Domain : constant String := WORM_Str.Get(String_Storage,X.Domain);
        Data_Last : natural := Data'last;
        Data_Length : natural := Data'length;
      begin
        if Domain'length > LC_Host'length-2
        or else Domain /= LC_Host(LC_Host'last - Domain'length+1 .. LC_Host'last) then
          return;
        end if;
        if Path'length > The_Path'length
        or else Path /= The_Path(The_Path'first .. The_Path'first+Path'length-1) then
          return;
        end if;
        -- Aha! it matches
        if X.Secure and not Secure then return;end if;
        if X.Expiration /= Unknown_Time
        and then X.Expiration < Ada.Calendar.Clock then
          return;  -- ignore expired cookie
        end if;
        if Last+Data'length > Result'last then -- truncate per spec
          Data_Length := Result'last-Last;
          if Data_Length < 3 then return;end if; -- no room for even "a=b"
          Data_Last := Data'first+Data_Length-1;
          if Ada.Strings.Fixed.Index(Data(Data'first .. Data_Last-1),"=")=0 then
            return;  -- can't fit even a single character to right of '='
          end if;
        end if;
        Result(Last+1 .. Last+Data_Length) := Data(Data'first .. Data_Last);
        Piece_Count := Piece_Count+1;
        Piece_List(Piece_Count):=(First=>Last+1,Last=>Last+Data_Length,
                                  Path_Length => Path'length);
        Last := Last+Data_Length;
      end Check_and_Build_Result;

    begin
      if Host'length < 2 or else Host(Host'first .. Host'first+1) /= "//" then
        return ""; -- illegal host name
      end if;
      for i in 1 .. Count loop
        Check_and_Build_Result(Jar(i));
      end loop;
      return Sorted_Result;
    end Get;

    function Take_Snapshot return Snapshot_Type is
    begin
      return (Count, Jar, String_Storage);
    end Take_Snapshot;

  end Cookie_Jar;

  procedure Put(Data : in String;
                Path : in String;
                Domain : in String;
                Expiration : in Ada.Calendar.Time := Unknown_Time;
                Secure : in Boolean) is
    -- ignores illegitimate data
    use Ada.Characters.Handling;
    Suffix : String(1 .. 4);
    Dot_Count : Natural := 0;
  begin
    if Path'length = 0 or else Path(Path'first) /= '/' then
      Put(Data, '/' & Path, Domain, Expiration, Secure);
      return;
    end if;
    if Domain'length < 3 then return;end if; -- illegal
    for i in Domain'range loop
      if Domain(i) = '.' then Dot_Count:=Dot_Count+1;end if;
    end loop;
    if Dot_Count < 2 then return;end if; -- illegal
    if Dot_Count = 2 then
      if Domain'length < 4 then
        return; -- illegal
      else
        Suffix := To_Lower(Domain(Domain'last-3 .. Domain'last));
        if Suffix /= ".com"
        and then Suffix /= ".edu"
        and then Suffix /= ".org"
        and then Suffix /= ".net"
        and then Suffix /= ".gov"
        and then Suffix /= ".mil"
        and then Suffix /= ".int" then
          return; -- illegal
        end if;
      end if;
    end if;
    Cookie_Jar.Put(Data, Path, Domain, Expiration, Secure);
  exception
    when others =>  -- problems? ignore it
      null;
  end Put;

  function Get(Host : String;
               Path : String;
               Secure : Boolean := False) return String is
  begin
    if Path'length = 0 or else Path(Path'first) /= '/' then
      return Get(Host, '/' & Path, Secure);
    end if;
    return Cookie_Jar.Get(Host,Path,Secure);
  end Get;

  function Take_Snapshot return Snapshot_Type is
  begin
    return Cookie_Jar.Take_Snapshot;
  end Take_Snapshot;

  function Count(Snapshot : Snapshot_Type) return Natural is
  begin return Snapshot.Count;end Count;

  function Data(Snapshot : Snapshot_Type; Index : Positive) return String is
  begin
    if Index > Snapshot.Count then raise Constraint_Error;end if;
    return WORM_Str.Get(Snapshot.Strings, Snapshot.Jar(Index).Data);
  end Data;

  function Path(Snapshot : Snapshot_Type; Index : Positive) return String is
  begin
    if Index > Snapshot.Count then raise Constraint_Error;end if;
    return WORM_Str.Get(Snapshot.Strings, Snapshot.Jar(Index).Path);
  end Path;

  function Domain(Snapshot : Snapshot_Type; Index : Positive) return String is
  begin
    if Index > Snapshot.Count then raise Constraint_Error;end if;
    return WORM_Str.Get(Snapshot.Strings, Snapshot.Jar(Index).Domain);
  end Domain;

  function Expiration(Snapshot : Snapshot_Type; Index : Positive)
    return Ada.Calendar.Time is
  begin
    if Index > Snapshot.Count then raise Constraint_Error;end if;
    return Snapshot.Jar(Index).Expiration;
  end Expiration;

  function Is_Secure(Snapshot : Snapshot_Type; Index : Positive) return Boolean is
  begin
    if Index > Snapshot.Count then raise Constraint_Error;end if;
    return Snapshot.Jar(Index).Secure;
  end Is_Secure;

  function Reconstruct(Snapshot : Snapshot_Type; Index : Positive) return String is
    function Cookie_Time(T : Ada.Calendar.Time) return String is
      use type Ada.Calendar.Time;
      use Claw.Time;
      S : String := Time_Image(T,
                               Date_Format=>Short_Month_Day_First,
                               Day_Of_Week_Format=>Short,
                               Time_Format=>Normal);
    begin
      if T = Unknown_Time then return "";end if;
      S(S'first+7) := '-';
      S(S'first+11) := '-';
      return "; expires=" & S & " GMT";
    end Cookie_Time;
    function Security(Secure : Boolean) return String is
    begin
      if Secure then return "; secure";else return "";end if;
    end Security;
  begin
    if Index > Snapshot.Count then raise Constraint_Error;end if;
    return WORM_Str.Get(Snapshot.Strings, Snapshot.Jar(Index).Data)
         & "; domain="
         & WORM_Str.Get(Snapshot.Strings, Snapshot.Jar(Index).Domain)
         & "; path="
         & WORM_Str.Get(Snapshot.Strings, Snapshot.Jar(Index).Path)
         & Cookie_Time(Snapshot.Jar(Index).Expiration)
         & Security(Snapshot.Jar(Index).Secure);
  end Reconstruct;

end Cookie;
