with Ada.Characters.Handling,
     Ada.Streams.Stream_IO,
     Ada.Unchecked_Deallocation,
     Claw.Directories,
     Claw.Sockets,
     Cookie;
package body HTTP is

  procedure Free is new Ada.Unchecked_Deallocation(String, String_Ptrs);

  procedure Open(Socket : in out Socket_Type;
                 Host   : in String;
                 Port   : in     Claw.Sockets.Port_Type := Claw.Sockets.http_port;
                 Patience:in     Duration := 60.0) is
    -- Raises Claw.Not_Found_Error if it can't find Host.
    use type Claw.Sockets.Error_Codes;
    use type Claw.Sockets.Network_Address_Type;
    Junk : Claw.Sockets.Port_Type;
  begin
    Socket.IP := Claw.Sockets.Read_As_Address(Host);
    if Socket.IP = 0 then
      Te_Text_Line_Sockets.Open(Te_Text_Line_Sockets.Socket_Type(Socket),
                                Domain_Name=>Host,
                                Port=>Port,
                                Timeout=>Patience);
      Get_Peer(Socket, Socket.IP, Junk);
    else
      Te_Text_Line_Sockets.Open(Te_Text_Line_Sockets.Socket_Type(Socket),
                                Address=>Socket.IP,
                                Port=>Port,
                                Timeout=>Patience);
    end if;
    Free(Socket.Host); -- should be nothing there
    Socket.Host := new String'(Host);
    Socket.Port := Port;
    Socket.Patience := Patience;
    Socket.HTTP_State := Did_Nothing;
    Socket.Error_Code := 0;
    Socket.Stream.Socket_Ptr := Socket'Unchecked_Access;
    Socket.Stream_Ptr := Socket.Stream'Unchecked_Access;
  exception
    when Claw.Windows_Error =>
      Te_Text_Line_Sockets.Get_Error_Code(
        Te_Text_Line_Sockets.Socket_Type(Socket),Socket.Error_Code);
      Close(Socket);
      if Socket.Error_Code = Claw.Sockets.WSA_Host_Not_Found
      or Socket.Error_Code = Claw.Sockets.WSAE_Connrefused
      or Socket.Error_Code = Claw.Sockets.WSAE_Timedout then
        raise Claw.Not_Found_Error;
      else
        raise;
      end if;
  end Open;

  procedure Open(Socket : in out Socket_Type;
                 IP     : in     Claw.Sockets.Network_Address_Type;
                 Host_Name : in  String; -- for "Host: www.a.b" in Get/etc
                 Port   : in     Claw.Sockets.Port_Type := Claw.Sockets.http_port;
                 Patience:in     Duration := 60.0) is
  begin
    Open(Socket, Claw.Sockets.String_Of(IP), Port, Patience);
    Free(Socket.Host); -- toss "1.2.3.4"
    Socket.Host := new String'(Host_Name); -- save "www.a.b"
  end Open;

  procedure Close(Socket : in out Socket_Type) is
    -- Closes the Socket.  Open is the sole legal next operation.
  begin
    Te_Text_Line_Sockets.Close(Te_Text_Line_Sockets.Socket_Type(Socket));
    Free(Socket.Host);
    Socket.HTTP_State := Did_Nothing;
    Socket.IP := 0;
    Socket.Stream.Socket_Ptr := null; -- kill the attached Stream
    Socket.Stream_Ptr := null;  -- make Stream(Socket) return null
    -- leave Error_Code untouched
  end Close;

  function IP_Address(Socket : Socket_Type) return Claw.Sockets.Network_Address_Type is
  -- raises Claw.Not_Valid_Error if the Socket isn't open.
    use type Claw.Sockets.Network_Address_Type;
  begin
    if Socket.IP = 0 then raise Claw.Not_Valid_Error;end if;
    return Socket.IP;
  end IP_Address;

  procedure Get_Error_Code(Socket : in out Socket_Type;
                           Error_Code :out Claw.Sockets.Error_Codes) is
  -- Note: Claw.Sockets.Image(Error_Code) returns a String.
  -- This call resets the saved error code to zero.
    use type Claw.Sockets.Error_Codes;
  begin
    Error_Code := Socket.Error_Code;
    Socket.Error_Code := 0;
    if Error_Code = 0 then -- go deeper
      Te_Text_Line_Sockets.Get_Error_Code(Te_Text_Line_Sockets.Socket_Type(Socket),
                                 Error_Code);
    end if;
  end Get_Error_Code;

  Mon : constant array (Ada.Calendar.Month_Number) of String(1 .. 3)
    := ("Jan", "Feb", "Mar", "Apr", "May", "Jun",
        "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");

  function Full_Date(Source : String) return Ada.Calendar.Time is
  -- scan an RFC 1123   Sun, 06 Nov 1994 08:49:37 GMT
  -- or 850(obsolete)   Sunday, 06-Nov-94 08:49:37 GMT
  -- Full Date and return as Ada.Calendar.Time
  -- raises Constraint_Error if problems
    use Ada.Calendar;
    Y2_Break: constant Integer := 70; -- -69 ->2069, -70 ->1970
    S       : constant String(1 .. Source'Length) := Source;
    Si      : Natural := 0;
    D       : Day_Number := Day_Number'First;
    M       : Month_Number := Month_Number'First;
    Y4      : Year_Number := Year_Number'First;
    Y2      : Integer range 0 .. 99 := 0;
    Hh      : Integer range 0 .. 23 := 0;
    Mm      : Integer range 0 .. 59 := 0;
    Ss      : Integer range 0 .. 59 := 0;
  begin
    if S(4) = ',' then -- Sun, 06 Nov 1994 08:49:37 GMT
      D := Ada.Calendar.Day_Number'Value(S(6 .. 7));
      for I in Month_Number loop
        if Mon(I) = S(9 .. 11) then
          M := I;
          exit;
        end if;
        if I = Month_Number'Last then
          raise Constraint_Error;
        end if;
      end loop;
      Y4 := Ada.Calendar.Year_Number'Value(S(13 .. 16));
      Hh := Integer'Value(S(18 .. 19));
      Mm := Integer'Value(S(21 .. 22));
      Ss := Integer'Value(S(24 .. 25));
    else -- Sunday, 06-Nov-94 08:49:37 GMT
      for I in S'range loop
        if S(I) = ' ' then
          Si := I + 1;
          exit;
        end if;
      end loop;
      D := Ada.Calendar.Day_Number'Value(S(Si .. Si + 1));
      for I in Ada.Calendar.Month_Number loop
        if Mon(I) = S(Si + 3 .. Si + 5) then
          M := I;
          exit;
        end if;
      end loop;
      Y2 := Ada.Calendar.Year_Number'Value(S(Si + 7 .. Si + 8));
      if Y2 < Y2_Break then
        Y4 := 2000 + Y2;
      else
        Y4 := 1900 + Y2;
      end if;
      Hh := Integer'Value(S(Si + 10 .. Si + 11));
      Mm := Integer'Value(S(Si + 13 .. Si + 14));
      Ss := Integer'Value(S(Si + 16 .. Si + 17));
    end if;
    return Time_Of(Y4, M, D, Day_Duration(Hh) * 3600
                             + Day_Duration(Mm) * 60
                             + Day_Duration(Ss));
  exception
    when others => return Claw.Directories.Unknown_Time;
  end Full_Date;

  procedure Scan_Header(Socket : in out Socket_Type;
                        Path : in String;
                        Response_Code : out Response_Code_Type) is
    use Ada.Characters.Handling;
    Seen_An_Info_Line: Boolean := False;
    Line : String(1 .. Socket.Max_Line_Length);
    Last : Natural;
  begin
    -- Path always starts with '/' because Action made sure
    loop
      -- scan returned header
      Te_Text_Line_Sockets.Get_Line(Te_Text_Line_Sockets.Socket_Type(Socket),
                                    Line, Last,
                                    Timeout => Socket.Patience);
      if Last = 0 then -- timeout
        if Seen_An_Info_Line then
          return;
        else
          raise Claw.Not_Found_Error;
        end if;
      end if;
      -- drop trailing [cr|lf]*lf
      if Last > 0 and then Line(Last) = Ascii.Lf then
        for I in reverse 1 .. Last loop
          if Line(I) = Ascii.Cr or Line(I) = Ascii.Lf then
            Last := I - 1;
          else
            exit;
          end if;
        end loop;
      end if;
      if Last = 0 then -- empty line
        if Seen_An_Info_Line then
          return;
        end if;
      else
        Seen_An_Info_Line := True;
      end if;
      -- handle different header response lines
      if Last >= 12 and then Line(1 .. 5) = "HTTP/" then
        Socket.HTTP_Version := Line(6 .. 8);
        if Line(6 .. 8) = "1.0" then
          Socket.Persistent_Connection := False;
        end if;
        Response_Code :=Line(10 .. 12);
      elsif Last > 10 and then Line(1 .. 10) = "Location: " then
        -- allow for prepended blanks
        for i in 11 .. Last loop
          if Line(i) /= ' ' then
            Socket.Alternate_Location
              := Alternate_Locations.To_Bounded_String(Line(i .. Last));
            exit;
          end if;
        end loop;
      elsif Last >= 12 and then Line(1 .. 12) = "Set-Cookie: " then
        Cookie.Put_String("//" & Socket.Host.all, Path, Line(13 .. Last));
      elsif Last >= 26
      and then To_Lower(Line(1 .. 26)) = "transfer-encoding: chunked" then
        Socket.Chunked_Transfer := True;
      elsif Last >= 17
      and then To_Lower(Line(1 .. 17)) = "connection: close" then
        Socket.Persistent_Connection := False;
      elsif Last > 15
      and then To_Lower(Line(1 .. 15)) = "content-length:" then
        Socket.Content_Length := Claw.DWord'value(Line(16 .. Last));
        Socket.Content_Length_Remaining := Socket.Content_Length;
      elsif Last > 15
      and then To_Lower(Line(1 .. 15)) = "last-modified: " then
        Socket.Last_Modified := Full_Date(Line(16 .. Last));
      elsif Last > 14
      and then To_Lower(Line(1 .. 14)) = "content-type: " then
        Socket.Content_Type
          := Content_Types.To_Bounded_String(Line(15 .. Last));
        declare
          di : constant Natural
            := Content_Types.Index(Socket.Content_Type,";");
        begin
          if di /= 0 then
            Content_Types.Head(Socket.Content_Type, di-1);
          end if;
        end;
      end if;
    end loop;
    -- never get here
  end Scan_Header;

  procedure Action(Method : in String;
                   Socket : in out Socket_Type;
                   Path   : in String;
                   Post_Data : in String;  -- non-null only for POST
                   Response_Code : out Response_Code_Type;
                   Send_Cookies : in Boolean;
                   Use_Proxy : in Boolean := False;
                   Content_Type_String : in String := "";
                   User_agent: in String;
                   Post_Length : in Claw.DWord :=0) is
    use type Claw.Sockets.Error_Codes;
    Partly_Handled_Windows_Error : exception;
    Try_Count : Integer;
  begin
    if not Is_Open(Socket) then
      raise Claw.Not_Valid_Error;
    end if;
    if Path'length = 0 or else Path(Path'first) /= '/' then
      Action(Method,Socket,'/' & Path,Post_Data,Response_Code,
             Send_Cookies,Use_Proxy,Content_Type_String,
             User_Agent,Post_Length);
      return;
    end if;
    if Socket.HTTP_State = Did_Nothing then
      Try_Count := 1;
    else
      -- We think connection is open because it's persistent.
      -- But if our attempt to interact finds the connection reset by the
      -- server, then we will have to re-open it and redo the interaction.
      Try_Count := 2;
      delay 0.1;  -- let server disconnect if he's gonna'
    end if;
    if Method = "HEAD" then
      Socket.HTTP_State := Did_Head;
    elsif Method = "GET" then
      Socket.HTTP_State := Did_Get;
    elsif Method = "POST" then
      Socket.HTTP_State := Did_Post;
    else
      raise Claw.Not_Valid_Error;
    end if;
    for Try in 1 .. Try_Count loop
      begin
        if Use_Proxy then
          -- Send a proxy request. Proxies often are too stupid to use
          -- HTTP 1.1. For HTTP 1.0, they need an entire path.
          Put_Line(Socket,
                   Method
                   & " http://"
                   & Socket.Host.all
                   & Path
                   & " HTTP/1.0");
          Socket.Persistent_Connection := False;  -- Since it's 1.0
          -- Department of redundancy department: the Wingate proxy is particularly
          -- stupid, and needs the host in the URL to find the server, but it
          -- doesn't send the host name to the server. So we repeat the Host line
          -- (that is what IE 4.0).
          -- domain is Host without initial //
          Put_Line(Socket, "Connection: close");
        else
          Put_Line(Socket,
                   Method & " " & Path & " HTTP/1.1");
        end if;
        Put_Line(Socket, "Host: " & Socket.Host.all);
        Put_Line(Socket, "User-agent: " & User_Agent);
        -- Be nice to the server operator, and explain ourselves.
        if Send_Cookies then
          declare
            Cookies : String := Cookie.Get(Socket.Host.all, Path);
          begin
            if Cookies'length /= 0 then
              Put_Line(Socket, "Cookie: " & Cookies);
            end if;
          end;
        end if;
        if Method = "POST" then
          Put_Line(Socket,"Content-type: " & Content_Type_String);
          Put_Line(Socket,"Content-length:" & Claw.DWord'image(Post_Length));
        end if;
        Put(Socket, Ascii.Cr & Ascii.Lf); -- empty line
        -- we've asked Server, now get his response
        -- Force to plain, in case this was a persistent socket used for chunked
        -- results previously.
        if Method = "POST" then
          Put(Socket, Post_Data);
        end if;
        Set_Encoding(Socket, Te_Text_Line_Sockets.Plain);
        Socket.HTTP_Version := "   ";
        Socket.Alternate_Location := Alternate_Locations.Null_Bounded_String;
        Socket.Content_Length := 0;
        Socket.Content_Type := Content_Types.Null_Bounded_String;
        Socket.Last_Modified := Claw.Directories.Unknown_Time;
        Socket.Persistent_Connection := True;
        Socket.Chunked_Transfer := False;
        Response_Code := "   ";
        Scan_Header(Socket, Path, Response_Code);

        exit; -- no reopen needed
      exception
        when Claw.Not_Found_Error =>
          if Try = Try_Count then raise;end if;  -- no more tries
          Te_Text_Line_Sockets.Close(Te_Text_Line_Sockets.Socket_Type(Socket));
          Te_Text_Line_Sockets.Open(Te_Text_Line_Sockets.Socket_Type(Socket),
                                    Address=>Socket.IP,
                                    Port=>Socket.Port,
                                    Timeout=>Socket.Patience);
          -- successful reopen, we can try again

        when Claw.Windows_Error =>
          if Try = Try_Count then raise;end if;  -- no more tries
          Te_Text_Line_Sockets.Get_Error_Code(
            Te_Text_Line_Sockets.Socket_Type(Socket), Socket.Error_Code);
          if Socket.Error_Code = Claw.Sockets.Wsae_Connreset then
            Te_Text_Line_Sockets.Close(Te_Text_Line_Sockets.Socket_Type(Socket));
            Te_Text_Line_Sockets.Open(Te_Text_Line_Sockets.Socket_Type(Socket),
                                      Address=>Socket.IP,
                                      Port=>Socket.Port,
                                      Timeout=>Socket.Patience);
            -- successful reopen, we can try again
          else
            raise Partly_Handled_Windows_Error;
          end if;
      end;
    end loop; -- of tries to Put action


    if Socket.Chunked_Transfer then
      Set_Encoding(Socket, Te_Text_Line_Sockets.Chunked);
    end if;
  exception
    when Claw.Not_Found_Error =>
      Close(Socket);
      raise;
    when Claw.Windows_Error =>
      Te_Text_Line_Sockets.Get_Error_Code(
        Te_Text_Line_Sockets.Socket_Type(Socket),Socket.Error_Code);
      Close(Socket);
      if Socket.Error_Code = Claw.Sockets.Wsae_Connreset then
        raise Claw.Not_Found_Error; -- other side has closed the socket
      end if;
      raise;
    when Partly_Handled_Windows_Error =>
      Close(Socket);
      raise Claw.Windows_Error;
    when others =>
      Close(Socket);
      raise;
  end Action;

  procedure Get(Socket : in out Socket_Type;
                Path   : in String;
                Response_Code : out Response_Code_Type;
                Send_Cookies : in Boolean := True;
                Use_Proxy : in Boolean := False;
                User_agent: String := "Ada HTTP") is
  begin
    Action("GET",Socket,Path,"",Response_Code,Send_Cookies,Use_Proxy,
           User_Agent=>User_Agent);
    Socket.HTTP_State := Did_Get;
  end Get;

  procedure Post(Socket : in out Socket_Type;
                 Path   : in String;
                 Content_Type : in String;
                 Data : in String;
                 Response_Code : out Response_Code_Type;
                 Send_Cookies : in Boolean := True;
                 Use_Proxy : in Boolean := False;
                 User_agent: String := "Ada HTTP") is
  begin
    Action("POST",Socket,Path,Data,Response_Code,Send_Cookies,Use_Proxy,
           Content_Type, User_Agent, Claw.DWord(Data'length));
    Socket.HTTP_State := Did_Post;
  end Post;

  procedure Post(Socket : in out Socket_Type;
                 Path   : in String;
                 Content_Type : in String;
                 Data : in Ada.Streams.Stream_Element_Array;
                 Response_Code : out Response_Code_Type;
                 Send_Cookies : in Boolean := True;
                 Use_Proxy : in Boolean := False;
                 User_agent: String := "Ada HTTP") is
    use type Ada.Streams.Stream_Element_Offset;
    si : Ada.Streams.Stream_Element_Offset := Data'first;
    SData : String(1 .. Data'length);
  begin
    for i in SData'range loop
      SData(i) := Character'val(Integer(Data(si)));
      si := si+1;
    end loop;
    Post(Socket,Path,Content_Type,SData,Response_Code,Send_Cookies,Use_Proxy,
         User_agent=>User_Agent);
  end Post;

  procedure Head(Socket : in out Socket_Type;
                 Path   : in String;
                 Response_Code : out Response_Code_Type;
                 Send_Cookies : in Boolean := True;
                 Use_Proxy : in Boolean := False;
                 User_agent: String := "Ada HTTP") is
  begin
    Action("HEAD",Socket,Path,"",Response_Code,Send_Cookies,Use_Proxy,
           User_Agent=>User_Agent);
    Socket.HTTP_State := Did_Head;
  end Head;

  function HTTP_Version(Socket : Socket_Type) return HTTP_Version_Type is
  begin
    if Socket.HTTP_State = Did_Nothing then
      raise Claw.Not_Valid_Error;
    end if;
    return Socket.HTTP_Version;
  end HTTP_Version;

  function Alternate_Location(Socket : Socket_Type) return String is
  begin
    if Socket.HTTP_State = Did_Nothing then
      raise Claw.Not_Valid_Error;
    end if;
    return Alternate_Locations.To_String(Socket.Alternate_Location);
  end Alternate_Location;

  function Content_Length(Socket : Socket_Type) return Claw.DWord is
  begin
    -- Content_Length returns zero if not known
    if Socket.HTTP_State = Did_Nothing then
      raise Claw.Not_Valid_Error;
    end if;
    return Socket.Content_Length;
  end Content_Length;

  function Content_Type(Socket : Socket_Type) return String is
  begin
    if Socket.HTTP_State = Did_Nothing then
      raise Claw.Not_Valid_Error;
    end if;
    return Content_Types.To_String(Socket.Content_Type);
  end Content_Type;

  function Last_Modified(Socket : Socket_Type) return Ada.Calendar.Time is
  begin
    -- Last_Modified returns Claw.Directories.Unknown_Time if not known.
    if Socket.HTTP_State = Did_Nothing then
      raise Claw.Not_Valid_Error;
    end if;
    return Socket.Last_Modified;
  end Last_Modified;

  function Persistent_Connection(Socket : Socket_Type) return Boolean is
  begin
    if Socket.HTTP_State = Did_Nothing then
      raise Claw.Not_Valid_Error;
    end if;
    return Socket.Persistent_Connection;
  end Persistent_Connection;

  function Says_Chunked_Transfer(Socket : Socket_Type) return Boolean is
  begin
    if Socket.HTTP_State = Did_Nothing then
      raise Claw.Not_Valid_Error;
    end if;
    return Socket.Chunked_Transfer;
  end Says_Chunked_Transfer;

  procedure Finalize(Socket : in out Socket_Type) is
  begin
    Socket.HTTP_State := Did_Nothing;
    Te_Text_Line_Sockets.Finalize(Te_Text_Line_Sockets.Socket_Type(Socket));
    Free(Socket.Host);
    Socket.Stream.Socket_Ptr := null; -- kill the attached Stream
    Socket.Stream_Ptr := null;  -- make Stream(Socket) return null
  end Finalize;

  function Stream(Socket : Socket_Type) return Stream_Access is
  begin
    if Socket.Stream_Ptr = null
    or Socket.Stream.Socket_Ptr = null then  -- orphan Stream or closed Socket
      raise Claw.Not_Valid_Error;
    end if;
    return Socket.Stream_Ptr;
  end Stream;

  procedure Read_Chunked(Socket : in out Socket_Type;
                 Timeout : in     Duration;
                 Item    :    out Ada.Streams.Stream_Element_Array;
                 Last    :    out Ada.Streams.Stream_Element_Offset) is
    use type Ada.Streams.Stream_Element_Offset;
    Line : String(1 .. Integer(Item'length));
    SLast: Natural;
    si : Ada.Streams.Stream_Element_Offset := Item'first;
  begin
    Te_Text_Line_Sockets.Get_Line(Te_Text_Line_Sockets.Socket_Type(Socket),
                                  Line, SLast, Timeout);
    for i in 1 .. SLast loop
      Item(si) := Ada.Streams.Stream_Element(Character'pos(Line(i)));
      si := si+1;
    end loop;
    Last := Ada.Streams.Stream_Element_Offset(SLast);
  end Read_Chunked;

  procedure Get_Line(Socket  : in out Socket_Type;
                     Line    :    out String;
                     Last    :    out Natural;
                     Timeout : in     Duration := 30.0) is
    use type Claw.DWord;
    To_Get : Claw.DWord := Claw.DWord(Line'length);
  begin
    if (not Socket.Chunked_Transfer)
    and then Socket.Content_Length > 0
    and then To_Get > Socket.Content_Length_Remaining then
      To_Get := Socket.Content_Length_Remaining;
    end if;
    Te_Text_Line_Sockets.Get_Line(Te_Text_Line_Sockets.Socket_Type(Socket),
                          Line(Line'first .. Line'first+Integer(To_Get)-1),
                          Last, Timeout);
    Socket.Content_Length_Remaining := Socket.Content_Length_Remaining
                                       - Claw.DWord(Last-Line'first+1);
  end Get_Line;

end HTTP;
