with Ada.Calendar,
     Ada.Streams,
     Ada.Strings.Bounded,
     Claw,
     Claw.Sockets,
     Te_Text_Line_Sockets;
package HTTP is

  type Socket_Type     -- (Max_Line_Length: Positive) is inherited
  is new Te_Text_Line_Sockets.Socket_Type with private;

  procedure Open(Socket : in out Socket_Type;
                 Host   : in     String;  -- eg "www.yahoo.com", "11.22.33.44"
                 Port   : in     Claw.Sockets.Port_Type := Claw.Sockets.http_port;
                 Patience:in     Duration := 60.0);
    -- Get/Head/Post/stream Read all share the same timeout value: Patience.
    -- Raises Claw.Not_Found_Error if it can't find Host.

  procedure Open(Socket : in out Socket_Type;
                 IP     : in     Claw.Sockets.Network_Address_Type;
                 Host_Name : in  String; -- for "Host: www.a.b" in Get/etc
                 Port   : in     Claw.Sockets.Port_Type := Claw.Sockets.http_port;
                 Patience:in     Duration := 60.0);

  procedure Close(Socket : in out Socket_Type);
    -- Closes the Socket.  Open is the sole legal next operation.

  procedure Get_Error_Code(Socket : in out Socket_Type;
                           Error_Code :out Claw.Sockets.Error_Codes);
  -- Note: Claw.Sockets.Image(Error_Code) returns a String.
  -- This call resets the saved error code to zero.

  subtype Response_Code_Type is String(1 .. 3);

  procedure Get(Socket : in out Socket_Type;
                Path   : in String;  -- eg "/index.htm", may include ?cgi info
                Response_Code : out Response_Code_Type;
                Send_Cookies : in Boolean := True;
                Use_Proxy : in Boolean := False;
                User_agent: String := "Ada HTTP");
    -- Initiate Get of a web page.  This routine just reads the header,
    -- leaving later Reads to get the actual data.
    -- "404" or "301" are legal results, but an attempt to read the
    -- non-existent data will raise Claw.Not_Found_Error.  Note that a page
    -- that is present, though it's text says "sorry, page not here",
    -- will appear to be a perfectly good, present, page.
    --
    -- Any cookies will always be saved in temporary storage.  They
    -- may be retrieved via Cookies.Get(Host, Path : String) return String.
    -- If there are cookies for this host & path, and Send_Cookies
    -- is True, they will automatically be sent.
    --
    -- Connections are persistent: either because the remote host said so,
    -- or, if it closed the socket, a new Get or Head will automatically
    -- re-open the connection.
    --
    -- Plain/Chunked encoding is set for future reads.
    --
    -- raises Claw.Not_Valid_Error if Socket is not open

--       following routines need full, not Intro version of Claw.Sockets
--procedure Get(Socket : in out Socket_Type;
--              Path   : in String;  -- eg "/index.htm", may include ?cgi info
--              Response_Code : out Response_Code_Type;
--              Result_File_Name : in String;
--              Send_Cookies : in Boolean := True;
--              Use_Proxy : in Boolean := False;
--              User_agent: String := "Ada HTTP");
-- -- Like Get, but downloads the web page if Response_Code(1) = '2'

  subtype HTTP_Version_Type is String(1 .. 3);

  function IP_Address(Socket : Socket_Type) return Claw.Sockets.Network_Address_Type;
  -- raises Claw.Not_Valid_Error if the Socket isn't open.

  -- The following all raise Claw.Not_Valid_Error if no Get/Head/Post
  -- has been done on the socket.

  function HTTP_Version(Socket : Socket_Type) return HTTP_Version_Type;
  function Alternate_Location(Socket : Socket_Type) return String;
  function Content_Length(Socket : Socket_Type) return Claw.DWord;
    -- Content_Length returns zero if not known
  function Content_Type(Socket : Socket_Type) return String;
  function Last_Modified(Socket : Socket_Type) return Ada.Calendar.Time;
    -- Last_Modified returns Claw.Directories.Unknown_Time if not known.
  function Persistent_Connection(Socket : Socket_Type) return Boolean;
  function Says_Chunked_Transfer(Socket : Socket_Type) return Boolean;

  procedure Head(Socket : in out Socket_Type;
                 Path   : in String;
                 Response_Code : out Response_Code_Type;
                 Send_Cookies : in Boolean := True;
                 Use_Proxy : in Boolean := False;
                 User_agent: String := "Ada HTTP");
    -- Just like Get with the same parameters except that a following Read will
    -- raise Claw.Not_Found_Error

  procedure Post(Socket : in out Socket_Type;
                 Path   : in String;
                 Content_Type : in String;  -- eg. "text/plain"
                 Data : in String;
                 Response_Code : out Response_Code_Type;
                 Send_Cookies : in Boolean := True;
                 Use_Proxy : in Boolean := False;
                 User_agent: String := "Ada HTTP");
    -- Sends Data of Content_Type, then like Get: gets returned header
    -- followed by any returned data.

  procedure Post(Socket : in out Socket_Type;
                 Path   : in String;
                 Content_Type : in String;  -- eg. "application/octet-stream"
                 Data : in Ada.Streams.Stream_Element_Array;
                 Response_Code : out Response_Code_Type;
                 Send_Cookies : in Boolean := True;
                 Use_Proxy : in Boolean := False;
                 User_agent: String := "Ada HTTP");
    -- Data is Stream_Element_Array instead of String

  type Stream_Type
    is tagged limited private;
  --  is new ada.streams.root_stream_type with private;  -- not with Janus 3.1.1d

--       following routines need full, not Intro version of Claw.Sockets
--procedure Read(Stream  : in out Stream_Type;
--               Item    :    out Ada.Streams.Stream_Element_Array;
--               Last    :    out Ada.Streams.Stream_Element_Offset);
--  -- raises Claw.Not_Valid_Error if the Stream's socket is not open
--  -- Obeys Chunked or Content_Length limitation for end of data.
--
--procedure Write(Stream  : in out Stream_Type;
--                Item    : in     Ada.Streams.Stream_Element_Array);
--  -- raises Claw.Not_Valid_Error if the Stream's socket is not open

  type Stream_Access is access all Stream_Type'Class;

  function Stream(Socket : Socket_Type) return Stream_Access;
    -- raises Claw.Not_Valid_Error if Socket is not open

  Non_Text  : Exception renames Te_Text_Line_Sockets.Non_Text;
    -- raised by Get_Line on apparent non-text data

  procedure Get_Line(Socket  : in out Socket_Type;
                     Line    :    out String;
                     Last    :    out Natural;
                     Timeout : in     Duration := 30.0);
    -- Return Line(Line'first .. Last)
    -- Input up through LF if LF occurs,
    -- Line(Line'range) if no LF seen,
    -- or Input through last received char if timeout or hangup or end of data.
    -- Obeys Chunked, or Content_Length limited, transfer.
    -- NOTE Timeout parameter is last (unlike Claw.Sockets.Get) for defaulting
    -- raises
    --   Claw.Not_Found_Error           if timeout or connection closed
    --   Te_Text_Line.Bad_Chunk         if chunked and screwed up
    --   Non_Text       if data has excessive fraction of "control" characters

  -- inherited:
  -- Note: Inherited routines are unaware of Content_Length or Chunked
  -- end of data, so they should not normally be used with persistent
  -- connections.  Use the routines above.
  --
  -- procedure Input (Socket : in out Socket_Type;
  --                  Item   :    out Ada.Streams.Stream_Element_Array;
  --                  Last   :    out Ada.Streams.Stream_Element_Offset);
  --
  -- procedure Input (Socket  : in out Socket_Type;
  --                  Timeout : in     Duration;
  --                  Item    :    out Ada.Streams.Stream_Element_Array;
  --                  Last    :    out Ada.Streams.Stream_Element_Offset);
  --
  -- procedure Input_All (Socket  : in out Socket_Type;
  --                      Timeout : in     Duration := 120.0;
  --                      Item    :    out Ada.Streams.Stream_Element_Array;
  --                      Last    :    out Ada.Streams.Stream_Element_Offset);

  -- The generic record IO routines are not inherited, but are available
  -- as  Text_Line_Socket.Read etc.
  -- Note: These routines are unaware of Content_Length or Chunked
  -- end of data, so they should not normally be used with persistent
  -- connections.
  -- generic
  --   type Element_Type is private;
  -- procedure Read (Socket : in out Socket_Type'class;
  --                 Item   :    out Element_Type);
  --
  -- generic
  --   type Element_Type is private;
  -- procedure Try_Read (Socket  : in out Socket_Type'class;
  --                     Timeout : in     Duration;
  --                     Item    :    out Element_Type);

private

  type States is (Did_Nothing, Did_Head, Did_Get, Did_Post);

  package Alternate_Locations
  is new Ada.Strings.Bounded.Generic_Bounded_Length(255);

  package Content_Types
  is new Ada.Strings.Bounded.Generic_Bounded_Length(80);

  type Socket_Access is access all Socket_Type;

  type Stream_Type
  is tagged limited record
  --  is new ada.streams.root_stream_type with record  --not with Janus 3.1.1d
    Socket_Ptr: Socket_Access;
  end record;

  type String_Ptrs is access all String;

  type Socket_Type     -- (Max_Line_Length: Positive) is inherited
  is new Te_Text_Line_Sockets.Socket_Type with record
    Stream  : aliased Stream_Type;
    Stream_Ptr: Stream_Access;
    Patience : Duration;
    HTTP_State : States := Did_Nothing;
    Host : String_Ptrs;
    Port : Claw.Sockets.Port_Type;
    IP : Claw.Sockets.Network_Address_Type := 0;
    Error_Code : Claw.Sockets.Error_Codes :=0;
    HTTP_Version: HTTP_Version_Type;
    Alternate_Location: Alternate_Locations.Bounded_String;
    Content_Length: Claw.DWord;
    Content_Length_Remaining: Claw.DWord;
    Content_Type: Content_Types.Bounded_String;
    Last_Modified: Ada.Calendar.Time;
    Persistent_Connection: Boolean;
    Chunked_Transfer: Boolean;
  end record;

  procedure Finalize(Socket : in out Socket_Type);

end HTTP;
