with Ada.Calendar,
     Ada.Strings.Fixed,
     Claw.Sockets,
     Claw.Time;
package body SMTP is

  procedure Get_A_Response(Socket : in out Socket_Type) is
    -- Get Response from server.  Return True iff Success.
    use type Ada.Calendar.Time;
    Quitting_Time : constant Ada.Calendar.Time
      := Ada.Calendar.Clock + Socket.Patience;
    Response: String(1 .. Socket.Max_Line_Length);
    R_Last  : Natural;
  begin
    Socket.Response_Code := "   ";
    loop
      Get_Line(Socket, Response, R_Last, Quitting_Time - Ada.Calendar.Clock);
      exit when R_Last <= 3 or else Response(4) /= '-';
      delay 0.1;
    end loop;
    if R_Last >= 1 then
      R_Last := Integer'min(R_Last, 3);
      Socket.Response_Code(1 .. R_Last) := Response(1 .. R_Last);
    end if;
  end Get_A_Response;

  procedure Open(Socket : in out Socket_Type;
                 Host   : in String;
                 Port   : in     Claw.Sockets.Port_Type := Claw.Sockets.mail_port;
                 Patience:in     Duration := 30.0) is
    -- Raises Claw.Not_Found_Error if it can't find Host.
    use type Claw.Sockets.Error_Codes;
    use type Claw.Sockets.Network_Address_Type;
    IP : Claw.Sockets.Network_Address_Type;
    Info : Claw.Sockets.Host_Info_Type;
  begin
    if Socket.Mail_State /= Closed then
      raise Claw.Already_Valid_Error;
    end if;
    Socket.Patience := Patience;
    Socket.Error_Code := 0;
    IP := Claw.Sockets.Read_As_Address(Host);
    if IP /= 0 then
      Text_Line_Sockets.Open(Text_Line_Sockets.Socket_Type(Socket),
                             Address=>IP,
                             Port=>Port,
                             Timeout=>Socket.Patience);
    else
      Text_Line_Sockets.Open(Text_Line_Sockets.Socket_Type(Socket),
                             Domain_Name=>Host,
                             Port=>Port,
                             Timeout=>Socket.Patience);
    end if;
    Get_A_Response(Socket);
    if Socket.Response_Code(1) /= '2' then
      Text_Line_Sockets.Close(Text_Line_Sockets.Socket_Type(Socket));
      raise Response_Error;
    end if;
    Claw.Sockets.Lookup_Host_Info ("", Info);
    Put_Line(Socket, "HELO " & Claw.Sockets.Host_Name(Info));
    Get_A_Response(Socket);
    if Socket.Response_Code(1) /= '2' then
      Text_Line_Sockets.Close(Text_Line_Sockets.Socket_Type(Socket));
      raise Response_Error;
    end if;
    Socket.Mail_State := Ready_To_Send_Header;
  exception
    when Claw.Windows_Error =>
      Text_Line_Sockets.Get_Error_Code(
        Text_Line_Sockets.Socket_Type(Socket),Socket.Error_Code);
      Close(Socket);
      if Socket.Error_Code = Claw.Sockets.WSA_Host_Not_Found
      or Socket.Error_Code = Claw.Sockets.WSAE_Connrefused
      or Socket.Error_Code = Claw.Sockets.WSAE_Timedout then
        raise Claw.Not_Found_Error;
      else
        raise;
      end if;
  end Open;

  procedure Open(Socket : in out Socket_Type;
                 IP     : in     Claw.Sockets.Network_Address_Type;
                 Host_Name : in  String; -- for "Host: www.a.b" in Get/etc
                 Port   : in     Claw.Sockets.Port_Type := Claw.Sockets.mail_port;
                 Patience:in     Duration := 30.0) is
  begin
    Open(Socket,
         Host => Claw.Sockets.String_Of(IP),
         Port=>Port, Patience=>Patience);
  end Open;

  procedure Get_Error_Code(Socket : in out Socket_Type;
                           Error_Code :out Claw.Sockets.Error_Codes) is
  -- Note: Claw.Sockets.Image(Error_Code) returns a String.
  -- This call resets the saved error code to zero.
    use type Claw.Sockets.Error_Codes;
  begin
    Error_Code := Socket.Error_Code;
    Socket.Error_Code := 0;
    if Error_Code = 0 then -- go deeper
      Text_Line_Sockets.Get_Error_Code(Text_Line_Sockets.Socket_Type(Socket),
                              Error_Code);
    end if;
  end Get_Error_Code;

  procedure Get_Response_Code(Socket : in out Socket_Type;
                              Response_Code : out Response_Code_Type) is
  begin
    Response_Code := Socket.Response_Code;
    Socket.Response_Code := "   ";
  end Get_Response_Code;

  Uniqueness_Char:character := 'A';
  function Message_Id return String is
    YYYY : Ada.Calendar.Year_Number;
    MM   : Ada.Calendar.Month_Number;
    DD   : Ada.Calendar.Day_Number;
    SS   : Ada.Calendar.Day_Duration;
    type Seconds is range 0 .. 86400;
    Sec  : Seconds;
    H, M, S: Natural;
  function Two_Digit(I : Integer) return String is
  begin
    if I < 10 then return '0' & Integer'image(I)(2);end if;
    return Integer'image(I)(2 .. 3);
  end Two_Digit;
  begin
    Ada.Calendar.Split(Ada.Calendar.Clock, YYYY,MM,DD,SS);
    Sec := Seconds(SS);
    H := Integer(Sec/3600);
    M := Integer((Sec-3600*Seconds(H))/60);
    S := Integer(Sec-3600*Seconds(H)-60*Seconds(M));
    if Uniqueness_Char = 'Z' then
      Uniqueness_Char := 'A';
    else
      Uniqueness_Char := Character'succ(Uniqueness_Char);
    end if;
    return Two_Digit(YYYY/100) & Two_Digit(YYYY mod 100)
      & Two_Digit(MM) & Two_Digit(DD)
      & Two_Digit(H) & Two_Digit(M) & Two_Digit(S)
      & Uniqueness_Char
      & "Claw.Mail";
  end Message_ID;

  procedure Send_Message_Info(Socket : in out Socket_Type;
                              Sender : in     String;
                              To     : in String;
                              CC, BCC, From, Reply_To : in String := "";
                              Subject : in String;
                              Header_Info_Continuing : Boolean := False) is
    procedure Send_To(List : in String) is
      a,z : natural;
    begin
      a := List'first;
      loop
        while a <= List'last loop
          exit when List(a) /= ' ';
          a := a+1;
        end loop;
        z := a+1;
        exit when z >= List'last;
        z := Ada.Strings.Fixed.Index(List(z .. List'last), ",");
        if z = 0 then
          z := List'last;
        else
          z := z-1;
        end if;
        Put_Line(Socket, "RCPT To:<" & List(a .. z) & '>');
        Get_A_Response(Socket);
        if Socket.Response_Code(1) /= '2' then
          Close(Socket);
          raise Response_Error;
        end if;
        a := z+2;
      end loop;
    end Send_To;

  begin
    if Socket.Mail_State /= Ready_To_Send_Header then
      raise Claw.Not_Valid_Error;
    end if;
    Put_Line(Socket, "MAIL FROM:<" & Sender & '>');
    Get_A_Response(Socket);
    if Socket.Response_Code(1) /= '2' then
      Close(Socket);
      raise Response_Error;
    end if;
    Send_To(To);
    Send_To(Cc);
    Send_To(Bcc);
    Put_Line(Socket, "Data");
    Get_A_Response(Socket);
    if Socket.Response_Code /= "354" then
      Close(Socket);
      raise Response_Error;
    end if;
    Put_Line(Socket, "To: " & To);
    if Cc /= "" then
      Put_Line(Socket, "cc: " & Cc);
    end if;
    if Bcc /= "" then
      Put_Line(Socket, "bcc: " & Bcc);
    end if;
    if From /= "" then
      Put_Line(Socket, "From: " & From);
    else
      Put_Line(Socket, "From: " & Sender);
    end if;
    if Reply_To /= "" then
      Put_Line(Socket, "Reply-To: " & Reply_To);
    end if;
    Put_Line(Socket, "Subject: " & Subject);
    Put_Line(Socket, "Date:"
      & Claw.Time.Time_Image(Claw.Time.To_UTC_Time(Ada.Calendar.Clock),
          Claw.Time.Short_Month_Day_First,
          Day_Of_Week_Format=>Claw.Time.None)
      & " GMT");
    Put_Line(Socket, "Message-ID: <" & Message_ID & '>');
    if not Header_Info_Continuing then
      Put_Line(Socket, "");
    end if;
    Socket.Mail_State := Header_Sent;
  end Send_Message_Info;

  procedure Put_Content_Line(Socket : in out Socket_Type;
                             Line : in String) is
  begin
    if Socket.Mail_State /= Header_Sent then
      raise Claw.Not_Valid_Error;
    end if;
    if Line(Line'first) = '.' then
      Put_Line(Socket, "." & Line);
    else
      Put_Line(Socket, Line);
    end if;
  end Put_Content_Line;

  procedure Finish_Message(Socket : in out Socket_Type) is
  begin
    if Socket.Mail_State /= Header_Sent then
      raise Claw.Not_Valid_Error;
    end if;
    Put_Line(Socket, ".");
    Get_A_Response(Socket);
    if Socket.Response_Code /= "250" then
      delay 0.1;
      Put_Line(Socket, "."); -- try again
      Get_A_Response(Socket);
      if Socket.Response_Code /= "250" then
        Close(Socket);
        raise Response_Error;
      end if;
    end if;
    Socket.Mail_State := Ready_To_Send_Header;
  end Finish_Message;

  procedure Close(Socket : in out Socket_Type) is
    -- Closes the Socket.  Open is the sole legal next operation.
  begin
    if Socket.Mail_State = Closed then return;
    elsif Socket.Mail_State = Header_Sent then
      Finish_Message(Socket);
    end if;
    Put_Line(Socket, "QUIT");
    Get_A_Response(Socket);
    Socket.Mail_State := Closed;
    Text_Line_Sockets.Close(Text_Line_Sockets.Socket_Type(Socket));
    -- leave Error_Code untouched
  end Close;

  procedure Finalize(Socket : in out Socket_Type) is
  begin
    Close(Socket);
    Text_Line_Sockets.Finalize(Text_Line_Sockets.Socket_Type(Socket));
  exception
    when others => null;
  end Finalize;

end SMTP;

--Sony Viao extended warranty
--330/3
--230/2
--165/1 yr
--   PCGFX370
--1-800-747-1170 Crystal Jackson
