with Ada.Finalization,
     Ada.Streams;
package WORM_Str is

  -- 11/17/02 -TWM- Clean version including I/O
  -- 12/19/02 -TWM- Make tagged so binary search version can extend it

  Overflow : exception;  -- Attempt to Add too many strings,
                         -- or too many characters of strings.
  Nonesuch : exception;  -- Attempt to Get non-existent ID.

  type String_Counts is range 0 .. 200_000;               -- arbitrary

  type String_Character_Counts is range 0 .. 200_000_000; -- arbitrary

  type Store_Type(Max_Strings : String_Counts;
                  Max_Characters : String_Character_Counts;
                  Task_Safe : Boolean)
    is new Ada.Finalization.Controlled with private;
    -- We root at Controlled to give flexibility to descendants
  -- takes 4*(Max_Strings+1)+Max_Characters bytes

  function Count(Store : Store_Type) return String_Counts;
  -- return number of stored strings.

  function Character_Count(Store : Store_Type) return String_Character_Counts;
  -- return number of stored characters.

  subtype String_Indices is String_Counts range 1 .. String_Counts'last;

  procedure Add(Store : in out Store_Type;
                S : in String;
                Index : out String_Indices);
  -- Add to store.  Return new index.
  -- Raise Overflow if too many or too long strings.

  function Get_Length(Store : Store_Type;
                      Index : String_Indices) return Natural;
  -- raise Nonesuch if no such string.

  function Get(Store : Store_Type;
               Index : String_Indices) return String;
  -- raise Nonesuch if no such string.

  procedure Store_Type_Write(Stream : access Ada.Streams.Root_Stream_Type'class;
                             Item   : in     Store_Type);
  -- for Store_Type'Write use Store_Type_Write;
  -- Does not write discriminants, only data.
  -- NOTE: This is not task safe and Item should not be modified while
  -- this is executing.

  procedure Store_Type_Read(Stream : access Ada.Streams.Root_Stream_Type'class;
                            Item   : out    Store_Type);
  -- for Store_Type'Read use Store_Type_Read;
  -- raises Constraint_Error if the discriminants of Item don't allow
  -- enough space for the stream's data.
  -- raises Ada.Streams.Data_Error if it hits EOF in the middle of an input.
  -- NOTE: This is not task safe and Item should not be accessed while
  -- this is executing.

  type Comparison_Map_Type is array(Character) of Character;

  procedure Comparison_Map_Type_Write(Stream : access Ada.Streams.Root_Stream_Type'class;
                                      Item   : in     Comparison_Map_Type);
  -- for Comparison_Map_Type'Write use Comparison_Map_Type_Write

  procedure Comparison_Map_Type_Read(Stream : access Ada.Streams.Root_Stream_Type'class;
                                     Item   : out    Comparison_Map_Type);
  -- for Comparison_Map_Type'Read use Comparison_Map_Type_Read

  function Case_Sensitive_Compare return Comparison_Map_Type;

  function Case_Insensitive_Compare return Comparison_Map_Type;

  function aAbBcC_Compare return Comparison_Map_Type;

  type Comparisons is (LT, EQ, GT);

  function Compare(Store : Store_Type;
                   Comparison_Map : Comparison_Map_Type;
                   Left, Right : String_Indices) return Comparisons;

  function Compare(Store : Store_Type;
                   Comparison_Map : Comparison_Map_Type;
                   Index : String_Indices; S : String) return Comparisons;

  generic
    type IDs is (<>);
    Storage : in out WORM_Str.Store_Type;
    Comparison : in WORM_Str.Comparison_Map_Type;
  package Ordered_Store is

    function Count return String_Counts;
    -- return the number of distinct strings in this instantiation.

    function Is_Present(S : in String) return Boolean;
    -- Just check if it's already present in store.

    function Put(S : String) return IDs;
    -- Add to store if it isn't there yet.  Return new or old ID.
    -- Raise Overflow if too many or too long strings.

    procedure Lookup(S : in String; Success : out Boolean; ID : out IDs);
    -- if S is not Present, Success := False
    -- else Success := True and ID is set.

    function Get_Length(ID : IDs) return Natural;

    function Get(ID : IDs) return String;
    -- raise Nonesuch if no such ID.

    type Order_Natural is range 0 .. String_Counts'last;
    subtype Order_Positive is Order_Natural range 1 .. Order_Natural'last;

    type Ordered_Lists is array(Order_Positive range <>) of IDs;

    function Get_Ordered_Contents return Ordered_Lists;
    -- returns array of all String_Indices in table, in order.

    procedure Write(Stream : access Ada.Streams.Root_Stream_Type'class);
    -- Writes the information for Ordered_Store EXCEPT parameters Storage
    -- and Comparison.  To output all the information needed to later recreate
    -- this Ordered_Store, write Storage (use WORM_Str.Store_Type_Write),
    -- then Comparison (Comparison_Map_Type_Write), then call this Write.
    -- If multiple Ordered_Stores share one Storage, just write Storage once,
    -- then, for each Ordered_Store, write Comparison and call this Write.
    -- NOTE: This is not task safe and the data should not be updated while
    -- this is executing.

    procedure Read(Stream : access Ada.Streams.Root_Stream_Type'class);
    -- Reads the information for Ordered_Store EXCEPT parameters Storage
    -- and Comparison.  Read them, then instantiate Ordered_Store with
    -- them, then call this Read to reload data.  If Ordered_Store is
    -- already instantiated (ie, you are replacing its data), ensure
    -- that the Comparison map does not change.
    -- NOTE: This is not task safe and the data should not be accessed while
    -- this is executing.

  end Ordered_Store;

private

  subtype Data_Indices is String_Character_Counts
    range 1 .. String_Character_Counts'last;

  type Data_Type is array(Data_Indices range <>) of Character;

  type Last_List_Type is array(String_Counts range <>) of Data_Indices;

  type Store_Type(Max_Strings : String_Counts;
                  Max_Characters : String_Character_Counts;
                  Task_Safe : Boolean)
      is new Ada.Finalization.Controlled with record
    Next_String : String_Indices := String_Indices'first;
    Last_List : Last_List_Type(0 .. Max_Strings);
    Data : Data_Type(1 .. Max_Characters);
  end record;

end WORM_Str;
