with Ada.Calendar,
     WORM_Str;
package Cookie is

  Max_Simultaneous_Cookies : constant := 300; -- FIFO after too many/big cookies
  Max_Cookie_String : constant := 4000;  -- max length of Get result

  procedure Put_String(Host : in String;
                       Path : in String;  --initial '/' may be explicit or implicit
                       Text : in String);
  -- from "Set-Cookie: Text"
  -- parses Text and calls Put (see below)
  -- Note: Host starts with "//"

  Unknown_Time : constant Ada.Calendar.Time
    := Ada.Calendar.Time_Of(1980, 1, 1, 0.0);

  procedure Put(Data : in String;
                Path : in String;
                Domain : in String;
                Expiration : in Ada.Calendar.Time := Unknown_Time;
                Secure : in Boolean);
  -- save cookie info
  -- ignores illegitimate data

  function Get(Host : String;
               Path : String;
               Secure : Boolean := False) return String;
  -- fetch any cookie info, return "" if none, otherwise returns
  -- concatenated "name1=value1; name2=value2;..."
  -- truncates name/value pair making result'length > Max_Cookie_String

  type Snapshot_Type is private;
  -- The set of cookies may be changing, so we provide a way to take
  -- a snapshot at a point in time, then query that snapshot for
  -- information about each cookie.  This could be used, for instance,
  -- for saving cookies on disk.

  function Take_Snapshot return Snapshot_Type;
  -- takes a snapshot of the current set of cookies

  function Count(Snapshot : Snapshot_Type) return Natural;
  -- returns number of cookies in the snapshot set

  function Data(Snapshot : Snapshot_Type; Index : Positive) return String;
  -- returns the name/value string for the Index-th cookie in the snapshot
  -- raises Constraint_Error if Index is out of range.

  function Path(Snapshot : Snapshot_Type; Index : Positive) return String;
  -- returns the path string for the Index-th cookie in the snapshot

  function Domain(Snapshot : Snapshot_Type; Index : Positive) return String;
  -- returns the domain string for the Index-th cookie in the snapshot

  function Expiration(Snapshot : Snapshot_Type; Index : Positive)
    return Ada.Calendar.Time;
  -- returns the expiration time for the Index-th cookie in the snapshot,
  -- or returns Unknown_Time if no expiration was specified

  function Is_Secure(Snapshot : Snapshot_Type; Index : Positive) return Boolean;
  -- returns True iff the Index-th cookie in the snapshot said "secure"

  function Reconstruct(Snapshot : Snapshot_Type; Index : Positive) return String;
  -- returns the reconstructed text that followed "Set-Cookie: "

private
  type Infos is record
    Data : WORM_Str.String_Indices;
    Path : WORM_Str.String_Indices;
    Domain : WORM_Str.String_Indices; -- in lower case for caseless matching
    Expiration : Ada.Calendar.Time;
    Secure : Boolean;
  end record;
  type Jar_Contents_Type is array(1 .. Max_Simultaneous_Cookies) of Infos;

  Max_Cookie_String_Count : constant := 3*Max_Simultaneous_Cookies;
  Max_Cookie_Character_Count : constant := 100*Max_Cookie_String_Count;

  subtype String_Storage_Type is WORM_Str.Store_Type(
              Max_Strings => Max_Cookie_String_Count,
              Max_Characters=> Max_Cookie_Character_Count,
              Task_Safe=>False); -- OK: only used inside protected Cookie_Jar

  type Snapshot_Type is record
    Count : Natural := 0;
    Jar : Jar_Contents_Type;
    Strings : String_Storage_Type;
  end record;

end Cookie;
