with Ada.Characters.Handling,
     Ada.Strings.Fixed,
     Ada.Unchecked_Conversion;
package body Text_Line_Sockets is

  Eol_Char : constant string(1 .. 1) := (1 => ascii.LF);

  procedure Get_Line(Socket  : in out Socket_Type;
                     Line    :    out String;
                     Last    :    out Natural;
                     Timeout : in     Duration := 30.0) is
    use Ada.Strings.Fixed;
    Si      : Natural;
    N_Avail,
    N_To_Get: Natural;

    procedure Get_More is
      Non_Text_Count : Natural := 0;
    begin
      Socket.Next := 1;
      Socket.Last := 0;
      Get(Socket, Timeout, Socket.Buffer, Socket.Last);
      if Socket.Last < Socket.Next then return;end if;
      for i in Socket.Next .. Socket.Last loop
        if Ada.Characters.Handling.Is_Control(Socket.Buffer(i))
        and then Socket.Buffer(i) not in Ascii.HT .. Ascii.CR then
          Non_Text_Count:=Non_Text_Count+1;
        end if;
      end loop;
      Socket.Non_Text_Fraction
        := (Socket.Non_Text_Fraction*(float(Socket.Buffer'length)
                                      -float(Socket.Last))
            +float(Non_Text_Count)*float(Socket.Last)
           )/float(Socket.Buffer'length)**2;
      if Socket.Non_Text_Fraction > 0.05 then
        raise Non_Text;
      end if;
    exception
      when We   : Claw.Windows_Error =>
        Socket.Last := 0;
        Socket.Status := Problem; -- next Get will fail
        Ada.Exceptions.Save_Occurrence (Source => we,
                                        Target => Socket.Occurrence);
      when Oops : Claw.Not_Found_Error =>
        -- socket has been closed, treat as timeout
        Socket.Status := EOF;
        Ada.Exceptions.Save_Occurrence (Source => Oops,
                                        Target => Socket.Occurrence);
    end Get_More;

  begin
    Last := Line'First - 1;
    case Socket.Status is
      when OK =>
        null;
      when EOF =>
        Socket.Status := OK; -- so next read will work (if socket re-opened)
        raise Claw.Not_Found_Error;
      when Problem =>
        Socket.Status := OK; -- to allow conversation to continue
        Ada.Exceptions.Reraise_Occurrence(Socket.Occurrence);
    end case;
    loop
      if Last >= Line'Last then
        return;
      end if;
      if Last >= Line'First and then Line(Last) = Ascii.Lf then
        return;
      end if;
      Si := Index(Socket.Buffer(Socket.Next .. Socket.Last), Eol_Char);
      if Si = 0 then
        N_Avail := Socket.Last - Socket.Next + 1;
      else
        N_Avail := Si - Socket.Next + 1;
      end if;
      N_To_Get := Integer'Min(N_Avail, Line'Last - Last);
      if N_To_Get > 0 then
        Line(Last + 1 .. Last + N_To_Get)
          := Socket.Buffer(Socket.Next .. Socket.Next + N_To_Get - 1);
        Socket.Next := Socket.Next + N_To_Get;
        Last := Last + N_To_Get;
      else
        -- n_avail=0 in socket buffer
        if Socket.Status = OK then
          Get_More;
          if Socket.Last = 0 then
            return; -- timeout or close, return with whatever we've got
          end if;
        else
          -- can't read any more, return with whatever we've got
          return;
        end if;
      end if;
    end loop;
  end Get_Line;

  procedure Close(Socket  : in out Socket_Type) is
  begin
    -- reset for next user (since we don't have an explicit Open)
    Socket.Last := 0;
    Socket.Next := 1;
    Socket.Status := OK;
    Socket.Non_Text_Fraction := 0.0;
    Claw.Sockets.Close(Claw.Sockets.Socket_Type(Socket));
  end Close;

  procedure Finalize(Socket : in out Socket_Type) is
  begin
    Close(Socket);
    Claw.Sockets.Finalize(Claw.Sockets.Socket_Type(Socket));
  exception
    when others => null;
  end Finalize;

end Text_Line_Sockets;
